@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *
- *   Copyright 1996-2020 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2022 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -131,8 +131,7 @@ struct optimization optimizing =
     { MAX_OPTIMIZE, OPTIM_ALL_ENABLED }; /* number of optimization passes to take */
 static int cmd_sb = 16;    /* by default */
 
-iflag_t cpu;
-static iflag_t cmd_cpu;
+iflag_t cpu, cmd_cpu;
 
 struct location location;
 bool in_absolute;                 /* Flag we are in ABSOLUTE seg */
@@ -143,9 +142,8 @@ static struct RAA *offsets;
 static struct SAA *forwrefs;    /* keep track of forward references */
 static const struct forwrefinfo *forwref;
 
-static const struct preproc_ops *preproc;
 static struct strlist *include_path;
-bool pp_noline;                 /* Ignore %line directives */
+static enum preproc_opt ppopt;
 
 #define OP_NORMAL           (1U << 0)
 #define OP_PREPROCESS       (1U << 1)
@@ -304,29 +302,29 @@ static void define_macros(void)
 
     if (oct->have_local) {
         strftime(temp, sizeof temp, "__?DATE?__=\"%Y-%m-%d\"", &oct->local);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
         strftime(temp, sizeof temp, "__?DATE_NUM?__=%Y%m%d", &oct->local);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
         strftime(temp, sizeof temp, "__?TIME?__=\"%H:%M:%S\"", &oct->local);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
         strftime(temp, sizeof temp, "__?TIME_NUM?__=%H%M%S", &oct->local);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
     }
 
     if (oct->have_gm) {
         strftime(temp, sizeof temp, "__?UTC_DATE?__=\"%Y-%m-%d\"", &oct->gm);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
         strftime(temp, sizeof temp, "__?UTC_DATE_NUM?__=%Y%m%d", &oct->gm);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
         strftime(temp, sizeof temp, "__?UTC_TIME?__=\"%H:%M:%S\"", &oct->gm);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
         strftime(temp, sizeof temp, "__?UTC_TIME_NUM?__=%H%M%S", &oct->gm);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
     }
 
     if (oct->have_posix) {
         snprintf(temp, sizeof temp, "__?POSIX_TIME?__=%"PRId64, oct->posix);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
     }
 
     /*
@@ -336,20 +334,20 @@ static void define_macros(void)
      */
     snprintf(temp, sizeof(temp), "__?OUTPUT_FORMAT?__=%s",
              ofmt_alias ? ofmt_alias->shortname : ofmt->shortname);
-    preproc->pre_define(temp);
+    pp_pre_define(temp);
 
     /*
      * Output-format specific macros.
      */
     if (ofmt->stdmac)
-        preproc->extra_stdmac(ofmt->stdmac);
+        pp_extra_stdmac(ofmt->stdmac);
 
     /*
      * Debug format, if any
      */
     if (dfmt != &null_debug_form) {
         snprintf(temp, sizeof(temp), "__?DEBUG_FORMAT?__=%s", dfmt->shortname);
-        preproc->pre_define(temp);
+        pp_pre_define(temp);
     }
 }
 
@@ -364,9 +362,9 @@ static void define_macros(void)
  */
 static void preproc_init(struct strlist *ipath)
 {
-    preproc->init();
+    pp_init(ppopt);
     define_macros();
-    preproc->include_path(ipath);
+    pp_include_path(ipath);
 }
 
 static void emit_dependencies(struct strlist *list)
@@ -526,8 +524,8 @@ int main(int argc, char **argv)
 
     timestamp();
 
-    iflag_set_default_cpu(&cpu);
-    iflag_set_default_cpu(&cmd_cpu);
+    set_cpu(NULL);
+    cmd_cpu = cpu;
 
     set_default_limits();
 
@@ -551,7 +549,6 @@ int main(int argc, char **argv)
     offsets = raa_init();
     forwrefs = saa_init((int32_t)sizeof(struct forwrefinfo));
 
-    preproc = &nasmpp;
     operating_mode = OP_NORMAL;
 
     parse_cmdline(argc, argv, 1);
@@ -576,6 +573,11 @@ int main(int argc, char **argv)
         }
     }
 
+    /* Have we enabled TASM mode? */
+    if (tasm_compatible_mode) {
+        ppopt |= PP_TASM;
+        nasm_ctype_tasm_mode();
+    }
     preproc_init(include_path);
 
     parse_cmdline(argc, argv, 2);
@@ -618,13 +620,13 @@ int main(int argc, char **argv)
             char *line;
 
             if (depend_missing_ok)
-                preproc->include_path(NULL);    /* "assume generated" */
+                pp_include_path(NULL);    /* "assume generated" */
 
-            preproc->reset(inname, PP_DEPS, depend_list);
+            pp_reset(inname, PP_DEPS, depend_list);
             ofile = NULL;
-            while ((line = preproc->getline()))
+            while ((line = pp_getline()))
                 nasm_free(line);
-            preproc->cleanup_pass();
+            pp_cleanup_pass();
             reset_warnings();
     } else if (operating_mode & OP_PREPROCESS) {
             char *line;
@@ -647,9 +649,9 @@ int main(int argc, char **argv)
             location.known = false;
 
             _pass_type = PASS_PREPROC;
-            preproc->reset(inname, PP_PREPROC, depend_list);
+            pp_reset(inname, PP_PREPROC, depend_list);
 
-            while ((line = preproc->getline())) {
+            while ((line = pp_getline())) {
                 /*
                  * We generate %line directives if needed for later programs
                  */
@@ -694,7 +696,7 @@ int main(int argc, char **argv)
 
             nasm_free(quoted_file_name);
 
-            preproc->cleanup_pass();
+            pp_cleanup_pass();
             reset_warnings();
             if (ofile)
                 fclose(ofile);
@@ -729,7 +731,7 @@ int main(int argc, char **argv)
         }
     }
 
-    preproc->cleanup_session();
+    pp_cleanup_session();
 
     if (depend_list && !terminate_after_phase)
         emit_dependencies(depend_list);
@@ -814,8 +816,7 @@ static char *quote_for_pmake(const char *str)
     }
 
     /* Convert N backslashes at the end of filename to 2N backslashes */
-    if (nbs)
-        n += nbs;
+    n += nbs;
 
     os = q = nasm_malloc(n);
 
@@ -824,10 +825,10 @@ static char *quote_for_pmake(const char *str)
         switch (*p) {
         case ' ':
         case '\t':
-            while (nbs--)
-                *q++ = '\\';
+            q = mempset(q, '\\', nbs);
             *q++ = '\\';
             *q++ = *p;
+            nbs = 0;
             break;
         case '$':
             *q++ = *p;
@@ -849,9 +850,8 @@ static char *quote_for_pmake(const char *str)
             break;
         }
     }
-    while (nbs--)
-        *q++ = '\\';
 
+    q = mempset(q, '\\', nbs);
     *q = '\0';
 
     return os;
@@ -1066,19 +1066,19 @@ static bool process_arg(char *p, char *q, int pass)
         case 'p':       /* pre-include */
         case 'P':
             if (pass == 2)
-                preproc->pre_include(param);
+                pp_pre_include(param);
             break;
 
         case 'd':       /* pre-define */
         case 'D':
             if (pass == 2)
-                preproc->pre_define(param);
+                pp_pre_define(param);
             break;
 
         case 'u':       /* un-define */
         case 'U':
             if (pass == 2)
-                preproc->pre_undefine(param);
+                pp_pre_undefine(param);
             break;
 
         case 'i':       /* include search path */
@@ -1142,10 +1142,8 @@ static bool process_arg(char *p, char *q, int pass)
             break;
 
         case 't':
-            if (pass == 2) {
+            if (pass == 1)
                 tasm_compatible_mode = true;
-                nasm_ctype_tasm_mode();
-            }
             break;
 
         case 'v':
@@ -1160,7 +1158,7 @@ static bool process_arg(char *p, char *q, int pass)
 
         case 'a':       /* assemble only - don't preprocess */
             if (pass == 1)
-                preproc = &preproc_nop;
+                ppopt |= PP_TRIVIAL;
             break;
 
         case 'w':
@@ -1311,15 +1309,15 @@ static bool process_arg(char *p, char *q, int pass)
                     break;
                 case OPT_INCLUDE:
                     if (pass == 2)
-                        preproc->pre_include(q);
+                        pp_pre_include(q);
                     break;
                 case OPT_PRAGMA:
                     if (pass == 2)
-                        preproc->pre_command("pragma", param);
+                        pp_pre_command("%pragma", param);
                     break;
                 case OPT_BEFORE:
                     if (pass == 2)
-                        preproc->pre_command(NULL, param);
+                        pp_pre_command(NULL, param);
                     break;
                 case OPT_LIMIT:
                     if (pass == 1)
@@ -1329,7 +1327,7 @@ static bool process_arg(char *p, char *q, int pass)
                     keep_all = true;
                     break;
                 case OPT_NO_LINE:
-                    pp_noline = true;
+                    ppopt |= PP_NOLINE;
                     break;
                 case OPT_DEBUG:
                     debug_nasm = param ? strtoul(param, NULL, 10) : debug_nasm+1;
@@ -1689,9 +1687,8 @@ static void assemble_file(const char *fname, struct strlist *depend_list)
         cpu = cmd_cpu;
         if (listname) {
             if (pass_final() || list_on_every_pass()) {
-                active_list_options = list_options;
                 lfmt->init(listname);
-            } else if (active_list_options) {
+            } else if (list_active()) {
                 /*
                  * Looks like we used the list engine on a previous pass,
                  * but now it is turned off, presumably via %pragma -p
@@ -1699,7 +1696,6 @@ static void assemble_file(const char *fname, struct strlist *depend_list)
                 lfmt->cleanup();
                 if (!keep_all)
                     remove(listname);
-                active_list_options = 0;
             }
         }
 
@@ -1716,11 +1712,11 @@ static void assemble_file(const char *fname, struct strlist *depend_list)
             location.known = true;
         ofmt->reset();
         switch_segment(ofmt->section(NULL, &globalbits));
-        preproc->reset(fname, PP_NORMAL, pass_final() ? depend_list : NULL);
+        pp_reset(fname, PP_NORMAL, pass_final() ? depend_list : NULL);
 
         globallineno = 0;
 
-        while ((line = preproc->getline())) {
+        while ((line = pp_getline())) {
             if (++globallineno > nasm_limit[LIMIT_LINES])
                 nasm_fatal("overall line count exceeds the maximum %"PRId64"\n",
                            nasm_limit[LIMIT_LINES]);
@@ -1740,9 +1736,9 @@ static void assemble_file(const char *fname, struct strlist *depend_list)
 
         end_of_line:
             nasm_free(line);
-        }                       /* end while (line = preproc->getline... */
+        }                       /* end while (line = pp_getline... */
 
-        preproc->cleanup_pass();
+        pp_cleanup_pass();
 
         /* We better not be having an error hold still... */
         nasm_assert(!errhold_stack);
@@ -1875,8 +1871,8 @@ static bool is_suppressed(errflags severity)
     if (!(warning_state[warn_index(severity)] & WARN_ST_ENABLED))
         return true;
 
-    if (preproc && !(severity & ERR_PP_LISTMACRO))
-        return preproc->suppress_error(severity);
+    if (!(severity & ERR_PP_LISTMACRO))
+        return pp_suppress_error(severity);
 
     return false;
 }
@@ -2109,8 +2105,7 @@ void nasm_verror(errflags severity, const char *fmt, va_list args)
         return;
 
     if (!(severity & (ERR_HERE|ERR_PP_LISTMACRO)))
-        if (preproc)
-            preproc->error_list_macros(severity);
+        pp_error_list_macros(severity);
 }
 
 /*
@@ -2155,8 +2150,7 @@ static void nasm_issue_error(struct nasm_errtext *et)
             here = where.filename ? " here" : " in an unknown location";
         }
 
-        if (warn_list && true_type < ERR_NONFATAL &&
-            !(pass_first() && (severity & ERR_PASS1))) {
+        if (warn_list && true_type < ERR_NONFATAL) {
             /*
              * Buffer up warnings until we either get an error
              * or we are on the code-generation pass.
@@ -2238,7 +2232,7 @@ static void help(FILE *out)
         "    -o outfile    write output to outfile\n"
         "    --keep-all    output files will not be removed even if an error happens\n"
         "\n"
-        "    -Xformat      specifiy error reporting format (gnu or vc)\n"
+        "    -Xformat      specify error reporting format (gnu or vc)\n"
         "    -s            redirect error messages to stdout\n"
         "    -Zfile        redirect error messages to file\n"
         "\n"
@@ -2268,7 +2262,7 @@ static void help(FILE *out)
         "       -Ld        show byte and repeat counts in decimal, not hex\n"
         "       -Le        show the preprocessed output\n"
         "       -Lf        ignore .nolist (force output)\n"
-        "       -Lm        show multi-line macro calls with expanded parmeters\n"
+        "       -Lm        show multi-line macro calls with expanded parameters\n"
         "       -Lp        output a list file every pass, in case of errors\n"
         "       -Ls        show all single-line macro definitions\n"
         "       -Lw        flush the output after every line (very slow!)\n"
