@@ -1,36 +1,50 @@
 #include "error.h"
 
-const char * const warning_name[39] = {
+const char * const warning_name[53] = {
 	NULL,
-	"bnd",
 	"db-empty",
-	"environment",
+	"ea-absolute",
+	"ea-dispsize",
 	"float-denorm",
 	"float-overflow",
 	"float-toolong",
 	"float-underflow",
-	"hle",
+	"forward",
 	"label-orphan",
 	"label-redef",
 	"label-redef-late",
-	"lock",
-	"macro-def-case-single",
-	"macro-def-greedy-single",
-	"macro-def-param-single",
-	"macro-defaults",
-	"macro-params-legacy",
-	"macro-params-multi",
-	"macro-params-single",
-	"negative-rep",
 	"number-overflow",
 	"obsolete-nop",
 	"obsolete-removed",
 	"obsolete-valid",
 	"phase",
+	"pp-else-elif",
+	"pp-else-else",
+	"pp-empty-braces",
+	"pp-environment",
+	"pp-macro-def-case-single",
+	"pp-macro-def-greedy-single",
+	"pp-macro-def-param-single",
+	"pp-macro-defaults",
+	"pp-macro-params-legacy",
+	"pp-macro-params-multi",
+	"pp-macro-params-single",
+	"pp-macro-redef-multi",
+	"pp-open-braces",
+	"pp-open-brackets",
+	"pp-open-string",
+	"pp-rep-negative",
+	"pp-sel-range",
+	"pp-trailing",
 	"pragma-bad",
 	"pragma-empty",
 	"pragma-na",
 	"pragma-unknown",
+	"prefix-bnd",
+	"prefix-hle",
+	"prefix-lock",
+	"prefix-opsize",
+	"prefix-seg",
 	"ptr",
 	"regsize",
 	"unknown-warning",
@@ -42,29 +56,32 @@ const char * const warning_name[39] = {
 	"all"
 };
 
-const struct warning_alias warning_alias[42] = {
+const struct warning_alias warning_alias[68] = {
 	{ "all",                      WARN_IDX_ALL },
 	{ "bad-pragma",               WARN_IDX_PRAGMA_BAD },
-	{ "bnd",                      WARN_IDX_BND },
+	{ "bnd",                      WARN_IDX_PREFIX_BND },
 	{ "db-empty",                 WARN_IDX_DB_EMPTY },
-	{ "environment",              WARN_IDX_ENVIRONMENT },
+	{ "ea-absolute",              WARN_IDX_EA_ABSOLUTE },
+	{ "ea-dispsize",              WARN_IDX_EA_DISPSIZE },
+	{ "environment",              WARN_IDX_PP_ENVIRONMENT },
 	{ "float-denorm",             WARN_IDX_FLOAT_DENORM },
 	{ "float-overflow",           WARN_IDX_FLOAT_OVERFLOW },
 	{ "float-toolong",            WARN_IDX_FLOAT_TOOLONG },
 	{ "float-underflow",          WARN_IDX_FLOAT_UNDERFLOW },
-	{ "hle",                      WARN_IDX_HLE },
+	{ "forward",                  WARN_IDX_FORWARD },
+	{ "hle",                      WARN_IDX_PREFIX_HLE },
 	{ "label-orphan",             WARN_IDX_LABEL_ORPHAN },
 	{ "label-redef",              WARN_IDX_LABEL_REDEF },
 	{ "label-redef-late",         WARN_IDX_LABEL_REDEF_LATE },
-	{ "lock",                     WARN_IDX_LOCK },
-	{ "macro-def-case-single",    WARN_IDX_MACRO_DEF_CASE_SINGLE },
-	{ "macro-def-greedy-single",  WARN_IDX_MACRO_DEF_GREEDY_SINGLE },
-	{ "macro-def-param-single",   WARN_IDX_MACRO_DEF_PARAM_SINGLE },
-	{ "macro-defaults",           WARN_IDX_MACRO_DEFAULTS },
-	{ "macro-params-legacy",      WARN_IDX_MACRO_PARAMS_LEGACY },
-	{ "macro-params-multi",       WARN_IDX_MACRO_PARAMS_MULTI },
-	{ "macro-params-single",      WARN_IDX_MACRO_PARAMS_SINGLE },
-	{ "negative-rep",             WARN_IDX_NEGATIVE_REP },
+	{ "lock",                     WARN_IDX_PREFIX_LOCK },
+	{ "macro-def-case-single",    WARN_IDX_PP_MACRO_DEF_CASE_SINGLE },
+	{ "macro-def-greedy-single",  WARN_IDX_PP_MACRO_DEF_GREEDY_SINGLE },
+	{ "macro-def-param-single",   WARN_IDX_PP_MACRO_DEF_PARAM_SINGLE },
+	{ "macro-defaults",           WARN_IDX_PP_MACRO_DEFAULTS },
+	{ "macro-params-legacy",      WARN_IDX_PP_MACRO_PARAMS_LEGACY },
+	{ "macro-params-multi",       WARN_IDX_PP_MACRO_PARAMS_MULTI },
+	{ "macro-params-single",      WARN_IDX_PP_MACRO_PARAMS_SINGLE },
+	{ "negative-rep",             WARN_IDX_PP_REP_NEGATIVE },
 	{ "not-my-pragma",            WARN_IDX_PRAGMA_NA },
 	{ "number-overflow",          WARN_IDX_NUMBER_OVERFLOW },
 	{ "obsolete-nop",             WARN_IDX_OBSOLETE_NOP },
@@ -73,10 +90,33 @@ const struct warning_alias warning_alias[42] = {
 	{ "orphan-labels",            WARN_IDX_LABEL_ORPHAN },
 	{ "other",                    WARN_IDX_OTHER },
 	{ "phase",                    WARN_IDX_PHASE },
+	{ "pp-else-elif",             WARN_IDX_PP_ELSE_ELIF },
+	{ "pp-else-else",             WARN_IDX_PP_ELSE_ELSE },
+	{ "pp-empty-braces",          WARN_IDX_PP_EMPTY_BRACES },
+	{ "pp-environment",           WARN_IDX_PP_ENVIRONMENT },
+	{ "pp-macro-def-case-single", WARN_IDX_PP_MACRO_DEF_CASE_SINGLE },
+	{ "pp-macro-def-greedy-single", WARN_IDX_PP_MACRO_DEF_GREEDY_SINGLE },
+	{ "pp-macro-def-param-single", WARN_IDX_PP_MACRO_DEF_PARAM_SINGLE },
+	{ "pp-macro-defaults",        WARN_IDX_PP_MACRO_DEFAULTS },
+	{ "pp-macro-params-legacy",   WARN_IDX_PP_MACRO_PARAMS_LEGACY },
+	{ "pp-macro-params-multi",    WARN_IDX_PP_MACRO_PARAMS_MULTI },
+	{ "pp-macro-params-single",   WARN_IDX_PP_MACRO_PARAMS_SINGLE },
+	{ "pp-macro-redef-multi",     WARN_IDX_PP_MACRO_REDEF_MULTI },
+	{ "pp-open-braces",           WARN_IDX_PP_OPEN_BRACES },
+	{ "pp-open-brackets",         WARN_IDX_PP_OPEN_BRACKETS },
+	{ "pp-open-string",           WARN_IDX_PP_OPEN_STRING },
+	{ "pp-rep-negative",          WARN_IDX_PP_REP_NEGATIVE },
+	{ "pp-sel-range",             WARN_IDX_PP_SEL_RANGE },
+	{ "pp-trailing",              WARN_IDX_PP_TRAILING },
 	{ "pragma-bad",               WARN_IDX_PRAGMA_BAD },
 	{ "pragma-empty",             WARN_IDX_PRAGMA_EMPTY },
 	{ "pragma-na",                WARN_IDX_PRAGMA_NA },
 	{ "pragma-unknown",           WARN_IDX_PRAGMA_UNKNOWN },
+	{ "prefix-bnd",               WARN_IDX_PREFIX_BND },
+	{ "prefix-hle",               WARN_IDX_PREFIX_HLE },
+	{ "prefix-lock",              WARN_IDX_PREFIX_LOCK },
+	{ "prefix-opsize",            WARN_IDX_PREFIX_OPSIZE },
+	{ "prefix-seg",               WARN_IDX_PREFIX_SEG },
 	{ "ptr",                      WARN_IDX_PTR },
 	{ "regsize",                  WARN_IDX_REGSIZE },
 	{ "unknown-pragma",           WARN_IDX_PRAGMA_UNKNOWN },
@@ -87,20 +127,28 @@ const struct warning_alias warning_alias[42] = {
 	{ "zext-reloc",               WARN_IDX_ZEXT_RELOC }
 };
 
-const char * const warning_help[39] = {
+const char * const warning_help[53] = {
 	NULL,
-	"invalid BND prefixes",
 	"no operand for data declaration",
-	"nonexistent environment variable",
+	"absolute address cannot be RIP-relative",
+	"displacement size ignored on absolute address",
 	"floating point denormal",
 	"floating point overflow",
 	"too many digits in floating-point number",
 	"floating point underflow",
-	"invalid HLE prefixes",
+	"forward reference may have unpredictable results",
 	"labels alone on lines without trailing `:\'",
 	"label redefined to an identical value",
 	"label (re)defined during code generation",
-	"LOCK prefix on unlockable instructions",
+	"numeric constant does not fit",
+	"instruction obsolete and is a noop on the target CPU",
+	"instruction obsolete and removed on the target CPU",
+	"instruction obsolete but valid on the target CPU",
+	"phase error during stabilization",
+	"%elif after %else",
+	"%else after %else",
+	"empty %{} construct",
+	"nonexistent environment variable",
 	"single-line macro defined both case sensitive and insensitive",
 	"single-line macro",
 	"single-line macro defined with and without parameters",
@@ -108,16 +156,22 @@ const char * const warning_help[39] = {
 	"improperly calling multi-line macro for legacy support",
 	"multi-line macro calls with wrong parameter count",
 	"single-line macro calls with wrong parameter count",
+	"redefining multi-line macro",
+	"unterminated %{...}",
+	"unterminated %[...]",
+	"unterminated string",
 	"regative %rep count",
-	"numeric constant does not fit",
-	"instruction obsolete and is a noop on the target CPU",
-	"instruction obsolete and removed on the target CPU",
-	"instruction obsolete but valid on the target CPU",
-	"phase error during stabilization",
+	"%sel() argument out of range",
+	"trailing garbage ignored",
 	"malformed %pragma",
 	"empty %pragma directive",
 	"%pragma not applicable to this compilation",
 	"unknown %pragma facility or directive",
+	"invalid BND prefix",
+	"invalid HLE prefix",
+	"LOCK prefix on unlockable instructions",
+	"invalid operand size prefix",
+	"segment prefix ignored in 64-bit mode",
 	"non-NASM keyword used in other assemblers",
 	"register size specification ignored",
 	"unknown warning in -W/-w or warning directive",
@@ -125,11 +179,11 @@ const char * const warning_help[39] = {
 	"warning stack empty",
 	"RESx in initialized section becomes zero",
 	"relocation zero-extended to match output format",
-	"any warning not specifially mentioned above",
+	"any warning not specifically mentioned above",
 	"all possible warnings"
 };
 
-const uint8_t warning_default[38] = {
+const uint8_t warning_default[52] = {
 	WARN_INIT_ON,
 	WARN_INIT_ON,
 	WARN_INIT_ON,
@@ -145,6 +199,14 @@ const uint8_t warning_default[38] = {
 	WARN_INIT_ON,
 	WARN_INIT_ON,
 	WARN_INIT_ON,
+	WARN_INIT_ON,
+	WARN_INIT_OFF,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
 	WARN_INIT_ERR,
 	WARN_INIT_ON,
 	WARN_INIT_ON,
@@ -155,11 +217,17 @@ const uint8_t warning_default[38] = {
 	WARN_INIT_ON,
 	WARN_INIT_ON,
 	WARN_INIT_ON,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
 	WARN_INIT_OFF,
 	WARN_INIT_OFF,
 	WARN_INIT_OFF,
 	WARN_INIT_OFF,
-	WARN_INIT_OFF,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
+	WARN_INIT_ON,
 	WARN_INIT_ON,
 	WARN_INIT_ON,
 	WARN_INIT_OFF,
@@ -170,4 +238,4 @@ const uint8_t warning_default[38] = {
 	WARN_INIT_ON
 };
 
-uint8_t warning_state[38];	/* Current state */
+uint8_t warning_state[52];	/* Current state */
