@@ -40,16 +40,16 @@
 
 #include "compiler.h"
 
-#include <stdio.h>
-#include <string.h>
-#include <inttypes.h>
 
 #define BUILD_DRIVERS_ARRAY
-#include "output/outform.h"
+#include "outform.h"
+#include "outlib.h"
 
-struct ofmt *ofmt_find(char *name, struct ofmt_alias **ofmt_alias)
+const struct ofmt *ofmt_find(const char *name,
+			     const struct ofmt_alias **ofmt_alias)
 {
-    struct ofmt **ofp, *of;
+    const struct ofmt * const *ofp;
+    const struct ofmt *of;
     unsigned int i;
 
     *ofmt_alias = NULL;
@@ -72,9 +72,10 @@ struct ofmt *ofmt_find(char *name, struct ofmt_alias **ofmt_alias)
     return NULL;
 }
 
-struct dfmt *dfmt_find(struct ofmt *ofmt, const char *name)
+const struct dfmt *dfmt_find(const struct ofmt *ofmt, const char *name)
 {
-    struct dfmt **dfp, *df;
+    const struct dfmt * const *dfp;
+    const struct dfmt *df;
 
     for (dfp = ofmt->debug_formats; (df = *dfp); dfp++) {
         if (!nasm_stricmp(name, df->shortname))
@@ -83,35 +84,52 @@ struct dfmt *dfmt_find(struct ofmt *ofmt, const char *name)
     return NULL;
 }
 
-void ofmt_list(struct ofmt *deffmt, FILE * fp)
+void ofmt_list(const struct ofmt *deffmt, FILE * fp)
 {
-    struct ofmt **ofp, *of;
+    const struct ofmt * const *ofp, *of;
     unsigned int i;
 
     /* primary targets first */
     for (ofp = drivers; (of = *ofp); ofp++) {
-        fprintf(fp, "  %c %-10s%s\n",
-                of == deffmt ? '*' : ' ',
-                of->shortname, of->fullname);
+        fprintf(fp, "       %-20s %s%s\n",
+                of->shortname,
+                of->fullname,
+                of == deffmt ? " [default]" : "");
     }
 
     /* lets walk through aliases then */
     for (i = 0; i < ARRAY_SIZE(ofmt_aliases); i++) {
         if (!ofmt_aliases[i].shortname)
             continue;
-        fprintf(fp, "    %-10s%s\n",
+        fprintf(fp, "       %-20s Legacy alias for \"%s\"\n",
                 ofmt_aliases[i].shortname,
-                ofmt_aliases[i].fullname);
+                ofmt_aliases[i].ofmt->shortname);
     }
 }
 
-void dfmt_list(struct ofmt *ofmt, FILE *fp)
+void dfmt_list(FILE *fp)
 {
-    struct dfmt **dfp, *df;
+    const struct ofmt * const *ofp;
+    const struct ofmt *of;
+    const struct dfmt * const *dfp;
+    const struct dfmt *df;
+    char prefixbuf[32];
+    const char *prefix;
 
-    for (dfp = ofmt->debug_formats; (df = *dfp); dfp++) {
-        fprintf(fp, "  %c %-10s%s\n",
-                df == dfmt ? '*' : ' ',
-                df->shortname, df->fullname);
+    for (ofp = drivers; (of = *ofp); ofp++) {
+        if (of->debug_formats && of->debug_formats != null_debug_arr) {
+            snprintf(prefixbuf, sizeof prefixbuf, "%s:",
+                     of->shortname);
+            prefix = prefixbuf;
+
+            for (dfp = of->debug_formats; (df = *dfp); dfp++) {
+                if (df != &null_debug_form)
+                    fprintf(fp, "       %-10s %-9s %s%s\n",
+                            prefix,
+                            df->shortname, df->fullname,
+                            df == of->default_dfmt ? " [default]" : "");
+                prefix = "";
+            }
+        }
     }
 }
