@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *
- *   Copyright 1996-2011 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2022 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -236,9 +236,6 @@
 #ifdef OF_NO_AS86
 #undef OF_AS86
 #endif
-#ifdef OF_NO_RDF2
-#undef OF_RDF2
-#endif
 #ifdef OF_NO_IEEE
 #undef OF_IEEE
 #endif
@@ -269,7 +266,6 @@ extern const struct ofmt of_as86;
 extern const struct ofmt of_obj;
 extern const struct ofmt of_win32;
 extern const struct ofmt of_win64;
-extern const struct ofmt of_rdf2;
 extern const struct ofmt of_ieee;
 extern const struct ofmt of_macho32;
 extern const struct ofmt of_macho64;
@@ -318,9 +314,6 @@ static const struct ofmt * const drivers[] = {
 #ifdef OF_WIN64
     &of_win64,
 #endif
-#ifdef OF_RDF2
-    &of_rdf2,
-#endif
 #ifdef OF_IEEE
     &of_ieee,
 #endif
